# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Field, Struct, FieldName, Collect
from nr.databind.json import JsonDefault, JsonSerializer, JsonMixin
from nr.interface import implements
from nr.pylang.utils import classdef
from nr.stream import Stream
from shore.core.plugins import (
  IBasePlugin,
  IBuildTarget,
  IPackagePlugin,
  IPublishTarget,
  IMonorepoPlugin,
  load_plugin,
  PluginNotFound)
from shore.mapper import mapper
from shore.util.ast import load_module_members
from shore.util.version import bump_version, Version
from typing import Any, Callable, Dict, Iterable, Optional, List, Type, Union
import ast
import collections
import copy
import logging
import os
import re
import shlex
import yaml

__all__ = ['Package', 'Monorepo']

logger = logging.getLogger(__name__)


class VersionSelector(object):
  """ A version selector string. """

  def __init__(self, selector):
    if isinstance(selector, VersionSelector):
      selector = selector._string
    self._string = selector.strip()

  def __str__(self):
    return str(self._string)

  def __repr__(self):
    return 'VersionSelector({!r})'.format(self._string)

  def __eq__(self, other):
    if type(self) == type(other):
      return self._string == other._string
    return False

  def __ne__(self, other):
    return not (self == other)

  def to_setuptools(self):  # type: () -> str
    """ Converts the version selector to a string that Setuptools/Pip can
    understand by expanding the `~` and `^` range selectors.

    Given a version number X.Y.Z, the selectors will be expanded as follows:

    - `^X.Y.Z` -> `>=X.Y.Z,<X+1.0.0`
    - `~X.Y.Z` -> `>=X.Y.Z,<X.Y+1.0`
    - `X.Y.Z -> ==X.Y.Z`
    """

    # Poor-mans test if this looks like the form 'X.Y.Z' without anything around it.
    if not ',' in self._string and self._string[0].isdigit():
      return '==' + self._string

    regex = r'[~^](\d+\.\d+(\.\d+)?[.\-\w]*)'
    def sub(match):
      index = {'^': 0, '~': 1}[match.group(0)[0]]
      max_version = match.group(1).split('.')[:3]
      if len(max_version) == 2:
        max_version.append('0')
      if '-' in max_version[-1]:
        max_version[-1] = max_version[-1].partition('-')[0]
      max_version[index] = str(int(max_version[index]) + 1)
      for i in range(index+1, 3):
        max_version[i] = '0'
      return '>={},<{}'.format(match.group(1), '.'.join(max_version))
    return re.sub(regex, sub, self._string)

  def is_semver_selector(self) -> bool:
    return self._string and self._string[0] in '^~' and ',' not in self._string

  def matches(self, version: Union[Version, str]) -> bool:
    if not self.is_semver_selector():
      # TODO (@NiklasRosenstein): Match setuptools version selectors.
      return False
    min_version = Version(self._string[1:])
    if self._string[0] == '^':
      max_version = bump_version(min_version, 'major')
    elif self._string[0] == '~':
      max_version = bump_version(min_version, 'minor')
    else:
      raise RuntimeError('invalid semver selector string {!r}'.format(self._string))
    return min_version <= Version(version) < max_version


VersionSelector.ANY = VersionSelector('*')


@JsonSerializer(deserialize='_deserialize')
class Requirement(object):
  """ A requirement is a combination of a package name and a version selector.
  """

  def __init__(self, package, version):  # type: (str, VersionSelector)
    if not isinstance(package, str):
      raise TypeError('expected str for package_name')
    if not isinstance(version, VersionSelector):
      raise TypeError('expected VersionSelector for version')
    self.package = package
    self.version = version

  def __str__(self):
    if self.version == VersionSelector.ANY:
      return self.package
    return '{} {}'.format(self.package, self.version)

  def __repr__(self):
    return repr(str(self))#'Requirement({!r})'.format(str(self))

  @classmethod
  def parse(cls, requirement_string):
    match = re.match(r'^\s*([^\s]+)(?:\s+(.+))?$', requirement_string)
    if not match:
      raise ValueError('invalid requirement: {!r}'.format(requirement_string))
    package, version = match.groups()
    return cls(package, VersionSelector(version or VersionSelector.ANY))

  def to_setuptools(self):  # type: () -> str
    if self.version == VersionSelector.ANY:
      return self.package
    return '{} {}'.format(self.package, self.version.to_setuptools())

  @classmethod
  def _deserialize(cls, mapper, node):
    if not isinstance(node.value, str):
      raise node.type_error()
    return Requirement.parse(node.value)


@JsonSerializer(deserialize='_deserialize')
class Requirements(object):
  """ Represents package requirements, consisting of a #RequirementsList *any*
  that is comprised of requirements that always need to be present, and
  additional #RequirementsList#s in *platforms* that depend on the platform
  or environment (eg. `linux`, `win32` or `test` may be the platform keys).

  Additionally, the dependency on `python` is stored as the extra *python*
  attribute.

  This class is deserialized the same that it is represented in memory.
  Example:

  ```yaml
  - python ^2.7|^3.4
  - nr.interface ^1.0.0
  - test:
    - pytest
    - PyYAML
  ```

  Results in a #Requirements object like

  ```
  Requirements(python=VersionSelector('^2.7|^3.4'), required=[
    Requirement('nr.interface ^1.0.0')], test=[Requirement('pytest'),
    Requirement('PyYAML')], platforms={})
  ```

  Attributes:
    python (Optional[VersionSelector]): A selector for the Python version.
    required (List[Requirement]): A list of requirements that always need
      to be installed for a package, no matter the environment.
    test (List[Requirement]): A list of requirements that need to be installed
      for testing.
    platforms (Dict[str, Requirements]): A mapping of platform names to
      the requirements that need to be installed in that environment.
      Environments are tested against `sys.platform` in the rendered setup
      file.
  """

  classdef.repr('python,required,platforms')

  def __init__(self):
    self.python = None
    self.required = []
    self.platforms = {}

  def __bool__(self):
    return bool(self.python or self.required or self.platforms)

  @classmethod
  def _deserialize(cls, mapper, node):
    deserialize_type = [(Requirement, dict)]
    items = mapper.deserialize_node(node.replace(datatype=deserialize_type))

    self = node.datatype.cls()
    for index, item in enumerate(items):
      self._extract_from_item(mapper, node.make_child(index, None, item))
    return self

  def _extract_from_item(self, mapper, node):
    item = node.value
    if isinstance(item, Requirement):
      if item.package == 'python':
        self.python = item.version
      else:
        self.required.append(item)
    elif isinstance(item, dict):
      if len(item) != 1:
        raise ValueError('expected only a single key in requirements list')
      for key, value in item.items():
        value = mapper.deserialize_node(node.make_child(key, [Requirement], value))
        if key in self.platforms:
          self.platforms[key].extend(value)
        else:
          self.platforms[key] = value


class RootRequirements(Requirements):

  classdef.repr(Requirements.__repr_properties__ + ['test', 'extra',])

  def __init__(self):
    super(RootRequirements, self).__init__()
    self.test = None
    self.extra = {}

  def _extract_from_item(self, mapper, node):
    item = node.value
    if isinstance(item, dict) and len(item) == 1 and 'extra' in item:
      values_node = node.make_child('extra', None, item['extra'])
      for key, value in item['extra'].items():
        self.extra[key] = mapper.deserialize_node(values_node.make_child(key, Requirements, value))
    elif isinstance(item, dict) and len(item) == 1 and 'test' in item:
      test_node = node.make_child('test', Requirements, item['test'])
      self.test = mapper.deserialize_node(test_node)
    else:
      super(RootRequirements, self)._extract_from_item(mapper, node)


@JsonSerializer(deserialize='_deserialize')
class Author(Struct):
  name = Field(str)
  email = Field(str)

  AUTHOR_EMAIL_REGEX = re.compile(r'([^<]+)<([^>]+)>')

  def __str__(self):
    return '{} <{}>'.format(self.name, self.email)

  @classmethod
  def _deserialize(cls, mapper, node):
    if isinstance(node.value, str):
      match = Author.AUTHOR_EMAIL_REGEX.match(node.value)
      if match:
        author = match.group(1).strip()
        email = match.group(2).strip()
        return Author(author, email)
    raise NotImplementedError


@JsonSerializer(deserialize='_deserialize')
class Datafile(Struct):
  """ Represents an entry in the #Package.datafiles configuration. Can be
  deserialized from a JSON-like object or a string formatted as
  `source:target,includepattern,!excludepattern`. """

  source = Field(str)
  target = Field(str, default='.')
  include = Field([str])
  exclude = Field([str])

  @classmethod
  def _deserialize(cls, mapper, node):
    if isinstance(node.value, str):
      left, patterns = node.value.partition(',')[::2]
      if ':' in left:
        source, target = left.partition(':')[::2]
      else:
        source, target = left, '.'
      if not source or not target:
        raise ValueError('invalid DataFile spec: {!r}'.format(node.value))
      include = []
      exclude = []
      for pattern in patterns.split(','):
        (exclude if pattern.startswith('!') else include).append(pattern.lstrip('!'))
      return Datafile(source, target, include, exclude)
    raise NotImplementedError


@JsonSerializer(deserialize='_deserialize')
class PluginConfig(Struct):
  name = Field(str)
  plugin = Field(IBasePlugin)

  @property
  def is_package_plugin(self) -> bool:
    return IPackagePlugin.provided_by(self.plugin)

  @property
  def is_monorepo_plugin(self) -> bool:
    return IMonorepoPlugin.provided_by(self.plugin)

  def get_checks(self, subject: 'BaseObject'):
    if isinstance(subject, Package) and self.is_package_plugin:
      logger.debug('getting package checks for plugin {}'.format(self.name))
      return self.plugin.check_package(subject)
    if isinstance(subject, Monorepo) and self.is_monorepo_plugin:
      logger.debug('getting monorepo checks for plugin {}'.format(self.name))
      return self.plugin.check_monorepo(subject)
    logger.debug('skipping plugin {}'.format(self.name))
    return ()

  def get_files(self, subject: 'BaseObject'):
    if isinstance(subject, Package) and self.is_package_plugin:
      logger.debug('getting package files for plugin {}'.format(self.name))
      return self.plugin.get_package_files(subject)
    if isinstance(subject, Monorepo) and self.is_monorepo_plugin:
      logger.debug('getting monorepo files for plugin {}'.format(self.name))
      return self.plugin.get_monorepo_files(subject)
    logger.debug('skipping plugin {}'.format(self.name))
    return ()

  def get_version_refs(self, subject: 'BaseObject'):
    if isinstance(subject, Package) and self.is_package_plugin:
      logger.debug('getting package version refs for plugin {}'.format(self.name))
      return self.plugin.get_package_version_refs(subject)
    if isinstance(subject, Monorepo) and self.is_monorepo_plugin:
      logger.debug('getting monorepo version refs for plugin {}'.format(self.name))
      return self.plugin.get_monorepo_version_refs(subject)
    logger.debug('skipping plugin {}'.format(self.name))
    return ()

  def get_build_targets(self, subject: 'BaseObject'):
    if isinstance(subject, Package) and self.is_package_plugin:
      logger.debug('getting package builders for plugin {}'.format(self.name))
      return self.plugin.get_package_build_targets(subject)
    logger.debug('skipping plugin {}'.format(self.name))
    return ()

  def get_publish_targets(self, subject: 'BaseObject'):
    if isinstance(subject, Package) and self.is_package_plugin:
      logger.debug('getting package publishers for plugin {}'.format(self.name))
      return self.plugin.get_package_publish_targets(subject)
    logger.debug('skipping plugin {}'.format(self.name))
    return ()

  @classmethod
  def _deserialize(cls, mapper, node):
    if isinstance(node.value, str):
      plugin_name = node.value
      config = None
    elif isinstance(node.value, dict):
      if 'type' not in node.value:
        node.value_error('missing "type" key')
      config = node.value.copy()
      plugin_name = config.pop('type')
    else:
      raise TypeError('expected str or dict')
    try:
      plugin_cls = load_plugin(plugin_name)
    except PluginNotFound as exc:
      raise ValueError('plugin "{}" not found'.format(exc))
    if plugin_cls.Config is not None and config is not None:
      config = mapper.deserialize_node(node.make_child(plugin_name, plugin_cls.Config, config))
    elif plugin_cls.Config is None and config:
      raise TypeError('plugin {} expects no configuration'.format(plugin_name))
    else:
      config = None
    return PluginConfig(plugin_name, plugin_cls.new_instance(config))


@JsonSerializer(deserialize='_deserialize')
class InstallHook(JsonMixin, Struct):
  event = Field(str, default=None)
  command = Field((str, [str]))  #: Can be a string or list of strings.

  def normalize(self) -> 'InstallHook':
    if isinstance(self.command, str):
      return InstallHook(self.event, shlex.split(self.command))
    return self

  @classmethod
  def _deserialize(cls, mapper, node):
    if isinstance(node.value, str):
      return InstallHook(None, node.value)
    elif isinstance(node.value, dict):
      if len(node.value) != 1:
        raise ValueError('expected only one key')
      event, command = next(iter(node.value.items()))
      command = mapper.deserialize_node(node.make_child('command', InstallHook.command.datatype, command))
      return InstallHook(event, command)
    else:
      raise NotImplementedError  # Default deserialization


class ObjectCache(object):
  """ Helper class for loading #Package or #Monorepo objects from files. It
  caches the loaded object so that the same is not loaded multiple times into
  separate instances. """

  def __init__(self):
    self._cache = {}

  def clear(self):
    self._cache.clear()

  def get_or_load(self, filename: str, load_func: Callable[[str], Any]) -> Any:
    filename = os.path.normpath(os.path.abspath(filename))
    if filename not in self._cache:
      self._cache[filename] = load_func(filename)
    return self._cache[filename]


class BaseObject(Struct):

  #: The name of the object (ie. package or repository name).
  name = Field(str)

  #: The version of the object.
  version = Field(Version)

  #: Plugins for this object.
  use = Field([PluginConfig])

  #: Directory where the "shore changelog" command stores the changelog
  #: YAML files.
  changelog_directory = Field(str, default='.changelog')

  #: A hidden attribute that is not deserialized but set during
  #: deserialization from file to know the file that the data was
  #: loaded from.
  filename = Field(str, default=None, hidden=True)

  #: Contains all the unhandled keys from the deserialization.
  unhandled_keys = Field([str], default=None, hidden=True)

  #: The cache that this object is stored in.
  cache = Field(ObjectCache, default=None, hidden=True)

  @property
  def directory(self) -> str:
    return os.path.dirname(self.filename)

  def has_plugin(self, plugin_name: str) -> bool:
    return any(x.name == plugin_name for x in self.use)

  def get_plugins(self) -> List[PluginConfig]:
    plugins = list(self.use)
    if not self.has_plugin('core'):
      core_plugin = load_plugin('core')
      plugins.insert(0, PluginConfig('core', core_plugin.new_instance(None)))
    return plugins

  def get_build_targets(self) -> Dict[str, IBuildTarget]:
    targets = {}
    for plugin in self.get_plugins():
      for target in plugin.get_build_targets(self):
        target_id = plugin.name + ':' + target.get_name()
        if target_id in targets:
          raise RuntimeError('build target ID {} is not unique'.format(target_id))
        targets[target_id] = target
    return targets

  def get_publish_targets(self) -> Dict[str, IPublishTarget]:
    targets = {}
    for plugin in self.get_plugins():
      for target in plugin.get_publish_targets(self):
        target_id = plugin.name + ':' + target.get_name()
        if target_id in targets:
          raise RuntimeError('publish target ID {} is not unique'.format(target_id))
        targets[target_id] = target
    return targets

  @classmethod
  def load(cls, filename: str, cache: ObjectCache) -> '_DeserializableFromFile':
    """ Deserializes *cls* from a YAML file specified by *filename*. """

    def _load(filename):
      collect = Collect()
      with open(filename) as fp:
        obj = mapper.deserialize(yaml.safe_load(fp), cls, filename=filename, decorations=[collect])
      obj.filename = filename
      obj.unhandled_keys = Stream.concat(
        (x.locator.append(k) for k in x.unknowns)
        for x in collect.nodes).collect()
      obj.cache = cache
      obj.on_load_hook()
      return obj

    return cache.get_or_load(filename, _load)

  def on_load_hook(self):
    """ Called after the object was loaded with #load(). """

    pass


class Monorepo(BaseObject):
  private = Field(bool, default=False)

  #: Overrides the version field as it's optional for monorepos.
  version = Field(Version, default=None)

  #: If this option is enabled, individual packages in the monorepo have
  #: no individual version number. The "version" field in the package.yaml
  #: must be consistent with the version of the monorepo. Bumping the version
  #: of the monorepo will automatically bump the version in all packages.
  #: Bumping the version of individual packages will fail.
  mono_versioning = Field(bool, FieldName('mono-versioning'), default=False)

  #: The use field is optional on monorepos.
  use = Field([PluginConfig], default=list)

  #: Plugins to be used for all packages in the monorepo (unless explicitly
  #: overwritten in the package.yaml file).
  packages_use = Field([PluginConfig], default=list)

  #: Fields that can be inherited by the package.
  author = Field(Author, default=None)
  license = Field(str, default=None)
  url = Field(str, default=None)
  tag_format = Field(str, FieldName('tag-format'), default='{version}')

  @property
  def local_name(self) -> str:
    return self.name

  def get_packages(self) -> Iterable['Package']:
    """ Loads the packages for this mono repository. """

    for name in os.listdir(self.directory):
      path = os.path.join(self.directory, name, 'package.yaml')
      if os.path.isfile(path):
        package = Package.load(path, self.cache)
        assert package.monorepo is self, "woah hold up"
        yield package

  def get_private(self) -> bool:
    return self.private

  def get_tag_format(self) -> str:
    return self.tag_format

  def get_tag(self, version: str) -> str:
    tag_format = self.get_tag_format()
    return tag_format.format(name=self.name, version=version)

  def get_build_targets(self) -> Dict[str, IBuildTarget]:
    """
    Returns the publish targets for the monorepo. This includes the targets for
    packages in the monorepo.
    """

    targets = super().get_build_targets()
    for package in self.get_packages():
      for key, publisher in package.get_build_targets().items():
        targets[package.name + ':' + key] = self._BuildTargetWrapper(package.name, publisher)
    return targets

  def get_publish_targets(self) -> Dict[str, IPublishTarget]:
    """
    Returns the publish targets for the monorepo. If #mono_versioning is enabled,
    this includes the targets of child packages.
    """

    targets = super().get_publish_targets()
    for package in self.get_packages():
      for key, publisher in package.get_publish_targets().items():
        targets[package.name + ':' + key] = self._PublishTargetWrapper(package.name, publisher)
    return targets

  @implements(IBuildTarget)
  class _BuildTargetWrapper:
    def __init__(self, prefix, target):
      self.prefix = prefix
      self.target = target
    def get_name(self):
      return self.prefix + ':' + self.target.get_name()
    def get_build_artifacts(self):
      return self.target.get_build_artifacts()
    def build(self, build_directory):
      return self.target.build(build_directory)

  @implements(IPublishTarget)
  class _PublishTargetWrapper:
    def __init__(self, prefix, target):
      self.prefix = prefix
      self.target = target
    def get_name(self):
      return self.prefix + ':' + self.target.get_name()
    def get_build_selectors(self):
      return [self.prefix + ':' + k for k in self.target.get_build_selectors()]
    def publish(self, builds, test, build_directory, skip_existing):
      return self.target.publish(builds, test, build_directory, skip_existing)


class Package(BaseObject):
  #: Filled with the Monorepo if the package is associated with one. A package
  #: is associated with a monorepo if the parent directory of it's own
  #: directory contains a `monorepo.yaml` file.
  monorepo = Field(Monorepo, default=None, hidden=True)

  #: A private package will be prevented from being published with the
  #: "shore publish" command.
  private = Field(bool, default=None)

  #: The version number of the package.
  version = Field(Version, default=None)

  #: The author of the package.
  author = Field(Author, default=None)

  #: The license of the package. If #private is set to True, this can be None
  #: without a check complaining about it.
  license = Field(str, default=None)

  #: The URL of the package (eg. the GitHub repository).
  url = Field(str, default=None)

  #: A format specified when tagging a version of the package. This defaults
  #: to `"{version}"`. If the package is a member of a monorepo, #get_tag_format()
  #: adds the package name as a prefix.
  tag_format = Field(str, FieldName('tag-format'), default='{version}')

  #: The package description.
  description = Field(str)

  #: The default "use" field is populated with setuptools and pypi.
  use = Field([PluginConfig], default=list)

  #: The long description of the package. If this is not defined, the
  #: setuptools plugin will load the README file.
  long_description = Field(str, FieldName('long-description'), default=None)

  #: The content type for the long description. If not specified, the
  #: setuptools plugin will base that on the suffix of the README file.
  long_description_content_type = Field(str,
    FieldName('long-description-content-type'), default=None)

  #: The name of the module (potentially as a dottet path for namespaced
  #: modules). This is used to find the entry file in #get_entry_file().
  #: If not specified, the package #name is used.
  modulename = Field(str, default=None)

  #: The directory for the source files.
  source_directory = Field(str, FieldName('source-directory'), default='src')

  #: The names of packages that should be excluded when installing the
  #: package. The setuptools plugin will automatically expand the names
  #: here to conform with what the #setuptools.find_packages() function
  #: expects (eg. 'test' is converted into 'test' and 'test.*').
  exclude_packages = Field([str], FieldName('exclude-packages'),
    default=lambda: ['test', 'docs'])

  #: The requirements for the package.
  requirements = Field(RootRequirements, default=RootRequirements)

  #: The entrypoints for the package. The structure here is the same as
  #: for #setuptools.setup().
  entrypoints = Field({"value_type": [str]}, default=dict)

  #: A list of datafile definitions.
  datafiles = Field([Datafile], default=list)

  #: A list of instructions to render in the MANIFEST.in file.
  manifest = Field([str], default=list)

  #: Hooks that will be executed on install events (install/develop).
  install_hooks = Field([InstallHook], FieldName('install-hooks'), default=list)

  #: List of classifiers for the package.
  classifiers = Field([str], default=list)

  #: List of keywords.
  keywords = Field([str], default=list)

  #: Set to true to indicate that the package is typed. This will render a
  #: "py.typed" file in the source directory and include it in the package
  #: data.
  typed = Field(bool, default=False)

  @property
  def local_name(self) -> str:
    if self.monorepo:
      relpath = os.path.relpath(self.directory, self.monorepo.directory)
      return os.path.normpath(relpath)
    return self.name

  def _get_inherited_field(self, field_name: str) -> Any:
    value = getattr(self, field_name)
    if value is None and self.monorepo and hasattr(self.monorepo, field_name):
      value = getattr(self.monorepo, field_name)
    return value

  def get_private(self) -> bool:
    return self._get_inherited_field('private')

  def get_modulename(self) -> str:
    return self.modulename or self.name

  def get_version(self) -> str:
    version: str = self._get_inherited_field('version')
    if version is None:
      raise RuntimeError('version is not set')
    return version

  def get_author(self) -> Optional[Author]:
    return self._get_inherited_field('author')

  def get_license(self) -> str:
    return self._get_inherited_field('license')

  def get_url(self) -> str:
    return self._get_inherited_field('url')

  def get_plugins(self) -> List[PluginConfig]:
    plugins = super().get_plugins()

    # Inherit only plugins from the monorepo that are not defined in the
    # package itself.
    if self.monorepo and self.monorepo.packages_use:
      plugins.extend(x for x in self.monorepo.packages_use
        if not self.has_plugin(x.name))

    # Make sure there exists a setuptools and pypi target.
    if not any(x.name == 'setuptools' for x in plugins):
      plugins.append(mapper.deserialize('setuptools', PluginConfig))
    if not self.get_private():
      if not any(x.name == 'pypi' for x in plugins):
        plugins.append(mapper.deserialize('pypi', PluginConfig))

    return plugins

  def get_tag_format(self) -> str:
    tag_format = self._get_inherited_field('tag_format')
    if self.monorepo and '{name}' not in tag_format:
      tag_format = '{name}@' + tag_format
    return tag_format

  def get_tag(self, version: str) -> str:
    tag_format = self.get_tag_format()
    return tag_format.format(name=self.name, version=version)

  def on_load_hook(self):
    """ Called when the package is loaded. Attempts to find the Monorepo that
    belongs to this package and load it. """

    monorepo_fn = os.path.join(os.path.dirname(self.directory), 'monorepo.yaml')
    if os.path.isfile(monorepo_fn):
      self.monorepo = Monorepo.load(monorepo_fn, self.cache)

  def get_entry_file(self) -> str:
    """ Returns the filename of the entry file that contains package metadata
    such as `__version__` and `__author__`. """

    name = self.get_modulename().replace('-', '_')
    parts = name.split('.')
    prefix = os.sep.join(parts[:-1])
    for filename in [parts[-1] + '.py', os.path.join(parts[-1], '__init__.py')]:
      filename = os.path.join(self.source_directory, prefix, filename)
      if os.path.isfile(os.path.join(self.directory, filename)):
        return filename
    raise ValueError('Entry file for package "{}" could not be determined'
                     .format(self.name))

  def is_single_module(self) -> bool:
    return (
      self.get_modulename().count('.') == 0 and
      os.path.basename(self.get_entry_file()) != '__init__.py')

  def get_entry_file_abs(self) -> str:
    return os.path.normpath(os.path.join(self.directory, self.get_entry_file()))

  def get_entry_directory(self) -> str:
    """
    Returns the package directory. If this package is distributed in module-only
    form, a #ValueError is raised.
    """

    entry_file = self.get_entry_file()
    dirname, basename = os.path.split(entry_file)
    if basename != '__init__.py':
      raise ValueError('this package is in module-only form')
    return dirname

  EntryMetadata = collections.namedtuple('EntryFileData', 'author,version')

  def get_entry_metadata(self) -> EntryMetadata:
    """ Loads the entry file (see #get_entry_file()) and parses it with the
    #abc module to retrieve the value of the `__author__` and `__version__`
    variables defined in the file. """

    # Load the package/version data from the entry file.
    entry_file = self.get_entry_file()
    members = load_module_members(os.path.join(self.directory, entry_file))

    author = None
    version = None

    if '__version__' in members:
      try:
        version = ast.literal_eval(members['__version__'])
      except ValueError as exc:
        version = '<Non-literal expression>'

    if '__author__' in members:
      try:
        author = ast.literal_eval(members['__author__'])
      except ValueError as exc:
        author = '<Non-literal expression>'

    return self.EntryMetadata(author, version)

  def is_universal(self) -> bool:
    """
    Checks if the package is a universal Python package (i.e. it is Python 2 and 3 compatible)
    by testing the `$.requirements.python` version selector. If none is specified, the package
    is also considered universal.
    """

    if not self.requirements.python:
      return True
    # TODO (@NiklasRosenstein): This method of detecting if the version selector
    #   selects a Python 2 and 3 version is very suboptimal.
    has_2 = re.search(r'\b2\b|\b2\.\b', str(self.requirements.python))
    has_3 = re.search(r'\b3\b|\b3\.\b', str(self.requirements.python))
    return bool(has_2 and has_3)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from fnmatch import fnmatch
from nr.proxy import proxy_decorator
from nr.stream import Stream
from shore import __version__
from shore.core.plugins import (
  CheckResult,
  FileToRender,
  IMonorepoPlugin,
  IPackagePlugin,
  VersionRef,
  write_to_disk)
from shore.mapper import mapper
from shore.model import Monorepo, ObjectCache, Package, VersionSelector
from shore.plugins.core import get_monorepo_interdependency_version_refs
from shore.util import git as _git
from shore.util.changelog import ChangelogEntry, ChangelogManager, render_changelogs
from shore.util.classifiers import get_classifiers
from shore.util.license import get_license_metadata, wrap_license_text
from shore.util.resources import walk_package_resources
from shore.util.version import get_commit_distance_version, parse_version, bump_version, Version
from termcolor import colored
from typing import Any, Dict, Iterable, List, Optional, Union
import click
import io
import jinja2
import json
import logging
import nr.fs
import os
import pkg_resources
import re
import shlex
import shutil
import subprocess
import sys
import textwrap
import yaml

_cache = ObjectCache()
logger = logging.getLogger(__name__)


def _get_author_info_from_git():
  try:
    name = subprocess.getoutput('git config user.name')
    email = subprocess.getoutput('git config user.email')
  except FileNotFoundError:
    return None
  if not name and not email:
    return None
  return '{} <{}>'.format(name, email)


def _commit_distance_version(subject: [Monorepo, Package]) -> Version:
  if isinstance(subject, Package) and subject.monorepo \
      and subject.monorepo.mono_versioning:
    subject = subject.monorepo
  return get_commit_distance_version(
    subject.directory,
    subject.version,
    subject.get_tag(subject.version)) or subject.version


def _editor_open(filename: str):
  editor = shlex.split(os.getenv('EDITOR', 'vim'))
  return subprocess.call(editor + [filename])


def _edit_text(text: str) -> str:
  """
  Opens an editor for the user to modify *text*.
  """

  with nr.fs.tempfile('.yml', dir=os.getcwd(), text=True) as fp:
    fp.write(text)
    fp.close()
    res = _editor_open(fp.name)
    if res != 0:
      sys.exit(res)
    with open(fp.name) as src:
      return src.read()


def _load_subject(allow_none: bool = False) -> Union[Monorepo, Package, None]:
  package, monorepo = None, None
  if os.path.isfile('package.yaml'):
    package = Package.load('package.yaml', _cache)
  if os.path.isfile('monorepo.yaml'):
    monorepo = Monorepo.load('monorepo.yaml', _cache)
  if package and monorepo:
    raise RuntimeError('found package.yaml and monorepo.yaml in the same '
      'directory')
  if not allow_none and not package and not monorepo:
    logger.error('no package.yaml or monorepo.yaml in current directory')
    exit(1)
  return package or monorepo


@click.group()
@click.option('-C', '--change-directory')
@click.option('-v', '--verbose', is_flag=True)
@click.version_option(version=__version__)
def cli(change_directory, verbose):
  logging.basicConfig(
    format='[%(levelname)s:%(name)s]: %(message)s' if verbose else '%(message)s',
    level=logging.DEBUG if verbose else logging.INFO)

  if change_directory:
    os.chdir(change_directory)


@cli.command()
@click.argument('output_type', type=click.Choice(['json', 'text', 'notice']))
@click.argument('license_name')
def license(output_type, license_name):
  """ Print license information, full text or short notice. """

  data = get_license_metadata(license_name)
  if output_type == 'json':
    print(json.dumps(data(), sort_keys=True))
  elif output_type == 'text':
    print(wrap_license_text(data['license_text']))
  elif ouutput_type == 'notice':
    print(wrap_license_text(data['standard_notice'] or data['license_text']))
  else:
    raise RuntimeError(output_type)


@cli.command('classifiers')
@click.argument('q', required=False)
def classifiers(q):
  """ Search for package classifiers on PyPI. """

  for classifier in get_classifiers():
    if not q or q.strip().lower() in classifier.lower():
      print(classifier)


@cli.command()
@click.argument('name')
@click.argument('directory', required=False)
@click.option('--author')
@click.option('--version')
@click.option('--license')
@click.option('--modulename')
@click.option('--monorepo', is_flag=True)
@click.option('--dry', is_flag=True)
@click.option('--force', '-f', is_flag=True)
def new(**args):
  """ Initialize a new project or repository. """

  if not args['directory']:
    args['directory'] = args['name']

  if not args['author']:
    args['author'] = _get_author_info_from_git()

  env_vars = {
    'name': args['name'],
    'version': args['version'],
    'author': args['author'],
    'license': args['license'],
    'modulename': args['modulename'],
    'name_on_disk': args['modulename'] or args['name'],
  }

  name_on_disk = args['modulename'] or args['name']

  def _render_template(template_string, **kwargs):
    assert isinstance(template_string, str), type(template_string)
    return jinja2.Template(template_string).render(**(kwargs or env_vars))

  def _render_file(fp, filename):
    content = pkg_resources.resource_string('shore', filename).decode()
    fp.write(_render_template(content))

  def _render_namespace_file(fp):
    fp.write("__path__ = __import__('pkgutil').extend_path(__path__, __name__)\n")

  def _get_template_files(template_path) -> Iterable[FileToRender]:
    # Render the template files to the target directory.
    for source_filename in walk_package_resources('shore', template_path):
      # Expand variables in the filename.
      name = name_on_disk.replace('-', '_').replace('.', '/')
      filename = _render_template(source_filename, name=name)
      dest = os.path.join(args['directory'], filename)
      yield FileToRender(
        None,
        os.path.normpath(dest),
        lambda _, fp: _render_file(fp, template_path + '/' + source_filename))

  def _get_package_files() -> Iterable[FileToRender]:
    yield from _get_template_files('templates/package')

    # Render namespace supporting files.
    parts = []
    for item in name_on_disk.replace('-', '_').split('.')[:-1]:
      parts.append(item)
      dest = os.path.join(args['directory'], 'src', *parts, '__init__.py')
      yield FileToRender(
        None,
        os.path.normpath(dest),
        lambda _, fp: _render_namespace_file(fp))
      dest = os.path.join(args['directory'], 'src', 'test', *parts, '__init__.py')
      yield FileToRender(
        None,
        os.path.normpath(dest),
        lambda _, fp: fp.write('pass\n'))

    # TODO (@NiklasRosenstein): Render the license file if it does not exist.

  def _get_monorepo_files() -> Iterable[FileToRender]:
    yield from _get_template_files('templates/monorepo')

  if args['monorepo']:
    files = _get_monorepo_files()
  else:
    files = _get_package_files()

  for file in files:
    if os.path.isfile(file.name) and not args['force']:
      print(colored('Skip ' + file.name, 'yellow'))
      continue
    print(colored('Write ' + file.name, 'cyan'))
    if not args['dry']:
      write_to_disk(file)


def _run_for_subject(subject: Union[Package, Monorepo], func) -> List[Any]:
  if isinstance(subject, Monorepo):
    subjects = [subject] + sorted(subject.get_packages(), key=lambda x: x.name)
    return [func(x) for x in subjects]
  else:
    return [func(subject)]


def _color_subject_name(subject: Union[Package, Monorepo]) -> str:
    color = 'blue' if isinstance(subject, Monorepo) else 'cyan'
    return colored(subject.name, color)


def _run_checks(subject, treat_warnings_as_errors: bool=False):
  def _collect_checks(subject):
    return Stream.concat(x.get_checks(subject) for x in subject.get_plugins())
  checks = Stream.concat(_run_for_subject(subject, _collect_checks)).collect()
  if not checks:
    logger.info('✔ no checks triggered')
    return True

  max_level = max(x.level for x in checks)
  if max_level == CheckResult.Level.INFO:
    status = 0
  elif max_level == CheckResult.Level.WARNING:
    status = 1 if treat_warnings_as_errors else 0
  elif max_level ==  CheckResult.Level.ERROR:
    status = 1
  else:
    assert False, max_level

  logger.info('%s %s check(s) triggered', '❌' if status != 0 else '✔',
    len(checks))

  colors = {'ERROR': 'red', 'WARNING': 'magenta', 'INFO': None}
  for check in checks:
    level = colored(check.level.name, colors[check.level.name])
    print('  {} ({}): {}'.format(level, _color_subject_name(check.on), check.message))

  logger.debug('exiting with status %s', status)
  return False


@cli.command('check')
@click.option('--treat-warnings-as-errors', is_flag=True)
def checks(treat_warnings_as_errors):
  """ Run checks. """

  subject = _load_subject()
  if not _run_checks(subject, treat_warnings_as_errors):
    exit(1)
  exit(0)


@cli.command('update')
@click.option('--skip-checks', is_flag=True)
@click.option('--dry', is_flag=True)
@click.option('--stage', is_flag=True, help='Stage changed files in Git.')
def update(skip_checks, dry, stage):
  """ (Re-)render files managed shore. """

  def _collect_files(subject):
    return Stream.concat(x.get_files(subject) for x in subject.get_plugins())

  subject = _load_subject()
  if not skip_checks:
    _run_checks(subject, True)

  files = _run_for_subject(subject, _collect_files)
  files = Stream.concat(files).collect()

  logger.info('⚪ rendering %s file(s)', len(files))
  for file in files:
    logger.info('  %s', os.path.relpath(file.name))
    if not dry:
      write_to_disk(file)

  if stage:
    _git.add([f.name for f in files])


@cli.command('verify')
@click.option('--tag', '-t', help='Specify the tag from CI checks to match with the tag produced by shore.')
@click.option('--tag-check', type=click.Choice(['require', 'if-present', 'skip', 'ignore']), default='if-present')
@click.option('--update-check', type=click.Choice(['require', 'skip', 'ignore']), default='require')
def verify(tag, tag_check, update_check):
  """ Check whether "update" would change any files. """

  def _virtual_update(subject) -> Iterable[str]:
    files = Stream.concat(x.get_files(subject) for x in subject.get_plugins())
    for file in files:
      if not os.path.isfile(file.name):
        yield file.name
        continue
      fp = io.StringIO()
      write_to_disk(file, fp=fp)
      with io.open(file.name, newline='') as on_disk:
        if fp.getvalue() != on_disk.read():
          yield file.name

  def _tag_matcher(subject) -> Iterable[Union[Monorepo, Package]]:
    if isinstance(subject, Monorepo) and not subject.mono_versioning:
      # Tagging workflows on mono-repos without mono-versioning are not supported.
      return; yield
    if subject.get_tag(subject.version) == tag:
      yield subject

  status = 0

  subject = _load_subject()

  if update_check != 'skip':
    files = _run_for_subject(subject, _virtual_update)
    files = Stream.concat(files).collect()
    if files:
      logger.warning('❌ %s file(s) would be changed by an update.', len(files))
      if update_check != 'ignore':
        status = 1
    else:
      logger.info('✔ no files would be changed by an update.')
    for file in files:
      logger.warning('  %s', os.path.relpath(file))

  if tag_check != 'skip':
    if tag_check == 'require' and not tag:
      logger.error('❌ the specified tag is an empty string')
      status = 1
    elif tag:
      matches = _run_for_subject(subject, _tag_matcher)
      matches = Stream.concat(matches).collect()
      if len(matches) == 0:
        # TODO (@NiklasRosenstein): If we matched the {name} portion of the
        #   tag_format (if present) we could find which package (or monorepo)
        #   the tag was intended for.
        logger.error('❌ tag %s did not match any of the available subjects', tag)
        if tag_check != 'ignore':
          status = 1
      elif len(matches) > 1:
        logger.error('❌ tag matches multiple subjects: %s', tag)
        for match in matches:
          logger.error('  %s', match.name)
        if tag_check != 'ignore':
          status = 1
      else:
        logger.info('✔ tag %s matches %s', tag, matches[0].name)

  exit(status)


def _get_version_refs(subject) -> List[VersionRef]:
  def _get(subject):
    for plugin in subject.get_plugins():
      yield plugin.get_version_refs(subject)

  if isinstance(subject, Monorepo) and subject.mono_versioning:
    version_refs = Stream.concat(_run_for_subject(subject, _get))
  else:
    version_refs = _get(subject)
  return Stream.concat(version_refs).collect()


@cli.command('bump')
@click.argument('version', required=False)
@click.option('--major', is_flag=True)
@click.option('--minor', is_flag=True)
@click.option('--patch', is_flag=True)
@click.option('--post', is_flag=True)
@click.option('--snapshot', is_flag=True)
@click.option('--tag', is_flag=True)
@click.option('--dry', is_flag=True)
@click.option('--skip-checks', is_flag=True)
@click.option('--force', '-f', is_flag=True)
@click.option('--allow-lower', is_flag=True)
@click.option('--push', is_flag=True)
@click.option('--update', is_flag=True)
@click.option('--publish')
def bump(**args):
  """ Bump version numbers. Either supply a target "version" (may require --force
  if the specified version is lower than the current) or specify one of the --major,
  --minor, --patch, --post or --snapshot flags.

  The "version" argument can also be one of the strings "major", "minor", "patch",
  "post" or "git" which is only for backwards compatibility and will be removed in a
  future version of shore.
  """

  subject = _load_subject()
  changelog_manager = ChangelogManager(subject.changelog_directory, mapper)

  bump_flags = ('major', 'minor', 'patch', 'post', 'snapshot')
  bump_args = ['--' + k for k in bump_flags if args[k]]
  if args['version']:
    bump_args.insert(0, '<version>')
  if len(bump_args) > 1:
    logger.error('incompatible arguments: ' + ', '.join(bump_args))
    exit(1)
  elif not bump_args:
    flags = ', '.join('--' + k for k in bump_flags)
    logger.error('missing arguments: specify a <version> or one of ' + flags)
    exit(1)

  # Warn for deprecated behavior.
  if args['version'] in ('post', 'patch', 'minor', 'major', 'git'):
    use_flag = '--' + args['version']
    if use_flag == '--git':
      use_flag = '--snapshot'
    logger.warning('Support for the %r argument is deprecated and will be removed in a '
      'future version of Shore. Please use the %s flag instead.', args['version'], use_flag)

  if not args['skip_checks']:
    _run_checks(subject, True)

  if args['push'] and not args['tag']:
    logger.error('--push needs --tag')
    exit(1)

  if isinstance(subject, Package) and subject.monorepo \
      and subject.monorepo.mono_versioning:
    if args['force']:
      logger.warning('forcing version bump on individual package version '
        'that is usually managed by the monorepo.')
    else:
      logger.error('cannot bump individual package version if managed by monorepo.')
      exit(1)

  version_refs = _get_version_refs(subject)
  if not version_refs:
    logger.error('no version refs found')
    exit(1)

  # Ensure the version is the same accross all refs.
  is_inconsistent = any(parse_version(x.value) != subject.version for x in version_refs)
  if is_inconsistent and not args['force']:
    logger.error('inconsistent versions across files need to be fixed first.')
    exit(1)
  elif is_inconsistent:
    logger.warning('inconsistent versions across files were found.')

  current_version = subject.version
  pep440_version = True
  if args['version'] == 'post' or args['post']:
    new_version = bump_version(current_version, 'post')
  elif args['version'] == 'patch' or args['patch']:
    new_version = bump_version(current_version, 'patch')
  elif args['version'] == 'minor' or args['minor']:
    new_version = bump_version(current_version, 'minor')
  elif args['version'] == 'major' or args['major']:
    new_version = bump_version(current_version, 'major')
  elif args['version'] == 'git' or args['snapshot']:
    new_version = _commit_distance_version(subject)
    args['allow_lower'] = True
  else:
    new_version = parse_version(args['version'])

  if not new_version.pep440_compliant:
    logger.warning('version "{}" is not PEP440 compliant.'.format(new_version))

  if new_version < current_version and not (args['force'] or args['allow_lower']):
    logger.error('version {} is lower than current version {}'.format(
      new_version, current_version))
    exit(1)
  # Comparing as strings to include the prerelease/build number in the
  # comparison.
  if str(new_version) == str(current_version) and not args['force']:
    logger.warning('new version {} is equal to current version {}'.format(
      new_version, current_version))
    exit(0)

  # The replacement below does not work if the same file is listed multiple
  # times so let's check for now that every file is listed only once.
  n_files = set(os.path.normpath(os.path.abspath(ref.filename))
                for ref in version_refs)
  assert len(n_files) == len(version_refs), "multiple version refs in one "\
    "file is not currently supported."

  logger.info('bumping %d version reference(s)', len(version_refs))
  for ref in version_refs:
    logger.info('  %s: %s → %s', os.path.relpath(ref.filename), ref.value, new_version)
    if not args['dry']:
      with open(ref.filename) as fp:
        contents = fp.read()
      contents = contents[:ref.start] + str(new_version) + contents[ref.end:]
      with open(ref.filename, 'w') as fp:
        fp.write(contents)

  # For monorepos using mono-versioning, we may need to bump cross-package references.
  if isinstance(subject, Monorepo) and subject.mono_versioning:
    version_sel_refs = list(get_monorepo_interdependency_version_refs(subject, new_version))
    logger.info('bumping %d monorepo inter-dependency requirement(s)', len(version_sel_refs))
    for group_key, refs in Stream.groupby(version_sel_refs, lambda r: r.filename, collect=list):
      logger.info('  %s:', os.path.relpath(group_key))
      with open(group_key) as fp:
        content = fp.read()
      offset = 0
      for ref in refs:
        logger.info('    %s %s → %s', ref.package, ref.sel, ref.new_sel)
        content = content[:ref.start - offset] + ref.new_sel + content[ref.end - offset:]
        offset += len(ref.sel) - len(ref.new_sel)
      if not args['dry']:
        with open(group_key, 'w') as fp:
          fp.write(content)

    if args['tag'] and version_sel_refs:
      logger.warning('bump requires an update in order to automatically tag')
      args['update'] = True

  changed_files = [x.filename for x in version_refs]

  # Rename the unreleased changelog if it exists.
  if changelog_manager.unreleased.exists():
    changed_files.append(changelog_manager.unreleased.filename)
    if args['dry']:
      changelog = changelog_manager.version(new_version)
    else:
      changelog = changelog_manager.release(new_version)
    changed_files.append(changelog.filename)
    logger.info('release staged changelog (%s → %s)', changelog_manager.unreleased.filename,
      changelog.filename)

  if args['update']:
    _cache.clear()
    try:
      update(['--stage'])
    except SystemExit as exc:
      if exc.code != 0:
        raise

  if args['tag']:
    if any(f.mode == 'A' for f in _git.porcelain()):
      logger.error('cannot tag with non-empty staging area')
      exit(1)

    tag_name = subject.get_tag(new_version)
    logger.info('tagging %s', tag_name)

    if not args['dry']:
      _git.add(changed_files)
      _git.commit('({}) bump version to {}'.format(subject.name, new_version), allow_empty=True)
      _git.tag(tag_name, force=args['force'])

    if not args['dry'] and args['push']:
      _git.push(_git.current_branch(), tag_name, force=args['force'])

  if args['publish']:
    _cache.clear()
    publish([args['publish']])


@cli.command('status')
def status():
  """ Print the release status. """

  subject = _load_subject()

  def _get_commits_since_last_tag(subject):
    tag = subject.get_tag(subject.version)
    ref = _git.rev_parse(tag)
    if not ref:
      return tag, None
    else:
      return tag, len(_git.rev_list(tag + '..HEAD', subject.directory))

  items = [subject]
  if isinstance(subject, Monorepo):
    items.extend(sorted(subject.get_packages(), key=lambda x: x.name))
    if not subject.version:
      items.remove(subject)
  width = max(len(x.local_name) for x in items)

  for item in items:
    tag, num_commits = _get_commits_since_last_tag(item)
    if num_commits is None:
      item_info = colored('tag "{}" not found'.format(tag), 'red')
    elif num_commits == 0:
      item_info = colored('no commits', 'green') + ' since "{}"'.format(tag)
    else:
      item_info = colored('{} commit(s)'.format(num_commits), 'yellow') + ' since "{}"'.format(tag)
    print('{}: {}'.format(item.local_name.rjust(width), item_info))


@cli.command()
@click.option('--tag', '-t', is_flag=True)
@click.option('--snapshot', '-s', is_flag=True)
def version(tag, snapshot):
  """ Print the current package or repository version. """

  subject = _load_subject()
  version = _commit_distance_version(subject) if snapshot else subject.version
  if tag:
    print(subject.get_tag(version))
  else:
    print(version)


@cli.command()
@click.argument('args', nargs=-1)
def git(args):
  """ Shortcut for running git commands with a version range since the last
  tag of the current package or repo.

  This is effectively a shortcut for

  \b
    git $1 `shore versions -ct`..HEAD $@ -- .
  """

  subject = _load_subject()
  tag = subject.get_tag(subject.version)
  command = ['git', args[0]] + [tag + '..HEAD'] + list(args[1:]) + ['--', '.']
  exit(subprocess.call(command))


def _filter_targets(targets: Dict[str, Any], target: str) -> Dict[str, Any]:
  result = {}
  for key, value in targets.items():
    if fnmatch(key, target) or fnmatch(key, target + ':*'):
      result[key] = value
  return result


@cli.command()
@click.argument('target')
@click.option('--build-dir', default='build',
  help='Override the build directory. Defaults to ./build')
def build(**args):
  """ Build distributions. """

  subject = _load_subject()
  targets = subject.get_build_targets()

  if args['target']:
    targets = _filter_targets(targets, args['target'])
    if not targets:
      logging.error('no build targets matched "%s"', args['target'])
      exit(1)

  if not targets:
    logging.info('no build targets')
    exit(0)

  os.makedirs(args['build_dir'], exist_ok=True)
  for target_id, target in targets.items():
    logger.info('building target %s', colored(target_id, 'cyan'))
    target.build(args['build_dir'])


@cli.command()
@click.argument('target', required=False)
@click.option('-l', '--list', is_flag=True)
@click.option('-a', '--all', is_flag=True)
@click.option('--build-dir', default='build',
  help='Override the build directory. Defaults to ./build')
@click.option('--test', is_flag=True,
  help='Publish to a test repository instead.')
@click.option('--build/--no-build', default=True,
  help='Always build artifacts before publishing. Enabled by default.')
@click.option('--skip-existing', is_flag=True)
def publish(**args):
  """ Publish a source distribution to PyPI. """

  subject = _load_subject()
  builds = subject.get_build_targets()
  publishers = subject.get_publish_targets()

  if args['all'] and isinstance(subject, Monorepo) and not subject.mono_versioning:
    logger.error('publish -a,--all not allowed for Monorepo without mono-versioning')
    exit(1)

  if args['target']:
    publishers = _filter_targets(publishers, args['target'])
    if not publishers:
      logger.error('no publish targets matched "%s"', args['target'])
      exit(1)

  if args['list']:
    if publishers:
      print('Publish targets for', colored(subject.name, 'cyan') + ':')
      for target in publishers:
        print('  ' + colored(target, 'yellow'))
    else:
      print('No publish targets for', colored(subject.name, 'cyan') + '.')
    exit(0)

  if not publishers or (not args['target'] and not args['all']):
    logging.info('no publish targets')
    exit(1)

  def _needs_build(build):
    for filename in build.get_build_artifacts():
      if not os.path.isfile(os.path.join(args['build_dir'], filename)):
        return True
    return False

  def _run_publisher(name, publisher):
    try:
      logging.info('collecting builds for "%s" ...', name)
      required_builds = {}
      for selector in publisher.get_build_selectors():
        selector_builds = _filter_targets(builds, selector)
        if not selector_builds:
          logger.error('selector "%s" could not be satisfied', selector)
          return False
        required_builds.update(selector_builds)

      for target_id, build in required_builds.items():
        if not args['build'] and not _needs_build(build):
          logger.info('skipping target %s', colored(target_id, 'cyan'))
        else:
          logger.info('building target %s', colored(target_id, 'cyan'))
          os.makedirs(args['build_dir'], exist_ok=True)
          build.build(args['build_dir'])

      publisher.publish(
        required_builds.values(),
        args['test'],
        args['build_dir'],
        args['skip_existing'])
      return True
    except:
      logger.exception('error while running publisher "%s"', name)
      return False

  status = 0
  for key, publisher in publishers.items():
    if not _run_publisher(key, publisher):
      status = 1

  logger.debug('exit with status code %s', status)
  exit(status)


@cli.command()
@click.argument('version', type=parse_version, required=False)
@click.option('--reformat', is_flag=True, help='reformat the changelog')
@click.option('--add', metavar='type,…', help='create a new changelog entry')
@click.option('--for', metavar='component,…', help='components for the new changelog entry')
@click.option('--fixes', metavar='issue,…', help='issues that this changelog entry fixes')
@click.option('-m', '--message', metavar='text', help='changelog entry description')
@click.option('-e', '--edit', is_flag=True, help='edit the changelog entry or file')
@click.option('--markdown', is_flag=True, help='render the changelog as markdown')
@click.option('-a', '--all', is_flag=True, help='show the changelog for all versions.')
def changelog(**args):
  """
  Show changelogs or create new entries.
  """

  if (args['version'] or args['reformat']) and args['add']:
    logger.error('unsupported combination of arguments')
    sys.exit(1)

  subject = _load_subject(allow_none=True)
  if subject:
    manager = ChangelogManager(subject.changelog_directory, mapper)
  else:
    manager = ChangelogManager(Package.changelog_directory.default, mapper)

  def _split(s: Optional[str]) -> List[str]:
    return list(filter(bool, map(str.strip, (s or '').split(','))))

  if args['add']:

    if not args['for']:
      args['for'] = 'general'

    # Warn about bad changelog types.
    for entry_type in _split(args['add']):
      if entry_type not in manager.TYPES:
        logger.warning('"%s" is not a well-known changelog entry type.', entry_type)

    entry = ChangelogEntry(
      _split(args['add']),
      _split(args['fixes']),
      _split(args['for']),
      args['message'] or '')

    # Allow the user to edit the entry if no description is provided or the
    # -e,--edit option was set.
    if not entry.description or args['edit']:
      serialized = yaml.safe_dump(mapper.serialize(entry, ChangelogEntry), sort_keys=False)
      entry = mapper.deserialize(yaml.safe_load(_edit_text(serialized)), ChangelogEntry)

    # Validate the entry contents (need a description and at least one type and component).
    if not entry.types or not entry.description or not entry.components:
      logger.error('changelog entries need at least one type and component and a description')
      sys.exit(1)

    created = not manager.unreleased.exists()
    manager.unreleased.add_entry(entry)
    manager.unreleased.save(create_directory=True)
    message = ('Created' if created else 'Updated') + ' "{}"'.format(manager.unreleased.filename)
    print(colored(message, 'cyan'))
    sys.exit(0)

  if args['edit']:
    if not manager.unreleased.exists():
      logger.error('no staged changelog')
      sys.exit(1)
    sys.exit(_editor_open(manager.unreleased.filename))

  changelogs = []
  if args['version'] or not args['all']:
    if args['all']:
      sys.exit('error: incompatible arguments: <version> and -a,--all')
    changelog = manager.version(args['version']) if args['version'] else manager.unreleased
    # Load the changelog for the specified version or the current staged entries.
    if not changelog.exists():
      print('No changelog for {}.'.format(colored(str(args['version'] or 'unreleased'), 'yellow')))
      sys.exit(0)
    changelogs.append(changelog)
  else:
    changelogs = list(manager.all())

  if args['reformat']:
    for changelog in changelogs:
      changelog.save()
    sys.exit(0)

  if args['markdown']:
    changelog_format = 'markdown'
  else:
    changelog_format = 'terminal'

  render_changelogs(sys.stdout, changelog_format, changelogs)


_entry_point = lambda: sys.exit(cli())

if __name__ == '__main__':
  _entry_point()



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Updates a conda-forge feedstock repository by cloning it, updating the version
details with data from PyPI and pushing it back to the feedstock. The GitHub PR
must still be created manually after this.

Todo:

\b
* Update run requirements and test imports
"""

from nr.interface import override, implements
from shore.__main__ import _load_subject
from termcolor import colored
from typing import List

import click
import logging
import nr.fs
import re
import requests
import subprocess
import sys

logger = logging.getLogger(__name__)


def _call(*args, **kwargs):
  proc = subprocess.Popen(*args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kwargs)
  stdout, _ = proc.communicate()
  if proc.returncode != 0:
    print(colored(stdout.decode(), 'red'))
    sys.exit(proc.returncode)


@click.command(help=__doc__)
@click.option('--update-feedstock', is_flag=True, help='Update the conda-forge feedstock for the current version.')
def main(update_feedstock):

  if not update_feedstock:
    logger.error('no operation specified')
    sys.exit(1)

  subject = _load_subject()

  repo_name = 'conda-forge/{}-feedstock'.format(subject.name)
  pypi_package_name = subject.name
  version = str(subject.version)
  clone_url = 'git@github.com:' + repo_name
  branch_name = 'v' + version

  s_repo_name = colored(repo_name, 'blue')
  s_pypi_package_name = colored(pypi_package_name, 'yellow')
  s_clone_url = colored(clone_url, 'blue')
  s_branch_name = colored(branch_name, 'yellow')

  print('Fetching PyPI record for', s_pypi_package_name, '...')
  url = 'https://pypi.org/pypi/{}/json'.format(pypi_package_name)
  pypi_data = requests.get(url).json()
  if version not in pypi_data['releases']:
    logger.error('No release on PyPI for %s v%s.', s_pypi_package_name, version)
    sys.exit(1)

  sdist = next((x for x in pypi_data['releases'][version] if x['packagetype'] == 'sdist'), None)
  if not sdist:
    logger.error('No sdist on PyPI for %s v%s.', s_pypi_package_name, version)
    sys.exit(1)
  sha256 = sdist['digests']['sha256']

  print('Updating', s_repo_name, 'from PyPI package', s_pypi_package_name + ' v' + version + '.')

  with nr.fs.tempdir() as tmpdir:

    print('Cloning', s_clone_url, '...')
    _call(['git', 'clone', clone_url, tmpdir.name])
    _call(['git', 'checkout', '-b', branch_name], cwd=tmpdir.name)

    recipe_path = nr.fs.join(tmpdir.name, 'recipe', 'meta.yaml')
    with open(recipe_path) as fp:
      recipe = fp.read()

    recipe = re.sub(r'(version = ")([^"]+)"', r'\g<1>' + version + '"', recipe)
    recipe = re.sub(r'(sha256:\s*")([^"]+)"', r'\g<1>' + sha256 + '"', recipe)

    with open(recipe_path, 'w') as fp:
      fp.write(recipe)

    print()
    subprocess.check_call(['git', '--no-pager', 'diff'], cwd=tmpdir.name)
    print()

    _call(['git', 'add', '.'], cwd=tmpdir.name)
    _call(['git', 'commit', '-m', 'Updating recipe for {} v{}.'.format(
      pypi_package_name, version)], cwd=tmpdir.name)

    print('Pushing branch', s_branch_name, 'to', s_clone_url, '...')
    _call(['git', 'push', 'origin', branch_name, '--set-upstream'], cwd=tmpdir.name)


if __name__ == '__main__':
  main()



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from pkg_resources import resource_string
from shore.core.plugins import FileToRender, IMonorepoPlugin
from shore.model import Monorepo, Package, Requirements
from shore.static import GENERATED_FILE_REMARK
from nr.databind.core import Field, Struct
from nr.interface import implements, override
from typing import Dict, Iterable, List
import networkx as nx
import json
import os


def _dirname(package: Package):
  return os.path.basename(package.directory)


@implements(IMonorepoPlugin)
class InstallScriptRenderer:
  """
  Renders a self-contained Python script for a monorepo that installs Pip packages
  in order, respecting their inter-dependencies.
  """

  class Config(Struct):
    filename = Field(str, default='bin/install')

  @override
  def get_monorepo_files(self, monorepo: Monorepo) -> Iterable[FileToRender]:
    graph = self._get_interpackage_dependencies(monorepo)
    pkg_order = list(nx.algorithms.dag.topological_sort(graph))
    package_def = '[\n'
    for pkgname in pkg_order:
      package = {
        'name': pkgname,
        'requires': graph.nodes[pkgname]['requires'],
        'extra_requires': graph.nodes[pkgname]['extra_requires']}
      package_def += '  ' + json.dumps(package, sort_keys=True) + ',\n'
    package_def += ']'

    def write_script(_current, fp):
      template = resource_string('shore', 'templates/install_script/install').decode('utf8')
      content = (template
        .replace('{{package_def}}', package_def)
        .replace('{{generated_file_remark}}', GENERATED_FILE_REMARK))
      fp.write(content)

    yield FileToRender(monorepo.directory,
      self.config.filename, write_script).with_chmod('+x')

  def _get_interpackage_dependencies(self, monorepo: Monorepo) -> nx.DiGraph:
    """
    Constructs a directed graph of the dependencies between the packages in
    *monorepo*. The nodes in the graph are the directory names of the packages.
    """

    packages = sorted(monorepo.get_packages(), key=lambda x: x.name, reverse=True)
    package_name_mapping = {}
    graph = nx.DiGraph()

    # Initialize graph nodes.
    for package in packages:
      node_id = _dirname(package)
      package_name_mapping[package.name] = node_id
      node = graph.add_node(
        node_id,
        directory=os.path.basename(package.directory),
        requires=[],
        extra_requires={})

    # Helper function to get just the deps that we need.
    def _flatten_reqs(dst: List[str], reqs: Requirements) -> List[str]:
      for req in reqs.required if reqs else ():
        if req.package in package_name_mapping:
          dst.append(package_name_mapping[req.package])
      return dst

    # Process requirements and create edges.
    for package in packages:
      node_id = package_name_mapping[package.name]
      extras = graph.nodes[node_id]['extra_requires']
      total_deps = _flatten_reqs([], package.requirements)
      graph.nodes[node_id]['requires'] += total_deps

      test = _flatten_reqs([], package.requirements.test)
      if test:
        total_deps += test
        extras['test'] = test

      for extra, reqs in package.requirements.extra.items():
        extra_reqs = _flatten_reqs([], reqs)
        if extra_reqs:
          total_deps += extra_reqs
          extras[extra] = extra_reqs

      [graph.add_edge(o, node_id) for o in total_deps]

    return graph



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Field, ObjectMapper, Struct
from nr.stream import Stream
from shore.util.version import Version
from termcolor import colored
from typing import Iterable, List, Optional, TextIO
import os
import re
import shutil
import textwrap
import yaml


class ChangelogEntry(Struct):
  types = Field([str])
  issues = Field([(str, int)], default=list)
  components = Field([str])
  description = Field(str)


class Changelog:

  RENDERERS = {}

  def __init__(self, filename: str, version: Optional[Version], mapper: ObjectMapper) -> None:
    self.filename = filename
    self.version = version
    self.mapper = mapper
    self.entries = []

  def exists(self) -> bool:
    return os.path.isfile(self.filename)

  def load(self) -> None:
    with open(self.filename) as fp:
      data = yaml.safe_load(fp)
    self.entries = self.mapper.deserialize(data, [ChangelogEntry], filename=self.filename)

  def save(self, create_directory: bool = False) -> None:
    if create_directory:
      os.makedirs(os.path.dirname(self.filename), exist_ok=True)
    data = self.mapper.serialize(self.entries, [ChangelogEntry])
    with open(self.filename, 'w') as fp:
      yaml.safe_dump(data, fp, sort_keys=False)

  def add_entry(self, entry: ChangelogEntry) -> None:
    self.entries.append(entry)


class ChangelogManager:

  TYPES = frozenset(['fix', 'improvement', 'docs', 'change', 'refactor', 'feature', 'enhancement'])

  def __init__(self, directory: str, mapper: ObjectMapper) -> None:
    self.directory = directory
    self.mapper = mapper
    self._cache = {}

  def _get(self, name: str, version: Optional[str]) -> Changelog:
    key = (name, str(version))
    if key in self._cache:
      return self._cache[key]
    changelog = Changelog(os.path.join(self.directory, name), version, self.mapper)
    if os.path.isfile(changelog.filename):
      changelog.load()
    self._cache[key] = changelog
    return changelog

  @property
  def unreleased(self) -> Changelog:
    return self._get('_unreleased.yml', None)

  def version(self, version: Version) -> Changelog:
    return self._get(str(version) + '.yml', version)

  def release(self, version: Version) -> Changelog:
    """
    Renames the unreleased changelog to the file name for the specified *version*.
    """

    unreleased = self.unreleased
    os.rename(unreleased.filename, self.version(version).filename)
    self._cache.clear()
    return self.version(version)

  def all(self) -> Iterable[Changelog]:
    """
    Yields all changelogs.
    """

    for name in os.listdir(self.directory):
      if not name.endswith('.yml'):
        continue
      if name == '_unreleased.yml':
        yield self.unreleased
      else:
        version = Version(name[:-4])
        yield self.version(version)


def _group_entries_by_component(entries):
  key = lambda x: x.components[0]
  return list(Stream.sortby(entries, key).groupby(key, collect=list))


def render_changelogs_for_terminal(fp: TextIO, changelogs: List[Changelog]) -> None:
  """
  Renders a #Changelog for the terminal to *fp*.
  """

  def _md_term_stylize(text: str) -> str:
    def _code(m):
      return colored(m.group(1), 'cyan')
    def _issue_ref(m):
      return colored(m.group(0), 'yellow', attrs=['bold'])
    text = re.sub(r'`([^`]+)`', _code, text)
    text = re.sub(r'#\d+', _issue_ref, text)
    return text

  def _fmt_issue(i):
    if str(i).isdigit():
      return '#' + str(i)
    return i

  def _fmt_issues(entry):
    if not entry.issues:
      return None
    return '(' + ', '.join(colored(_fmt_issue(i), 'yellow', attrs=['underline']) for i in entry.issues) + ')'

  def _fmt_types(entry):
    return ', '.join(colored(f, attrs=['bold']) for f in entry.types)

  def _fmt_components(entry):
    if len(entry.components) <= 1:
      return None
    return '(' + ', '.join(colored(f, 'red', attrs=['bold', 'underline']) for f in entry.components[1:]) + ')'

  if hasattr(shutil, 'get_terminal_size'):
    width = shutil.get_terminal_size((80, 23))[0]
  else:
    width = 80

  # Explode entries by component.
  for changelog in changelogs:
    fp.write(colored(changelog.version or 'Unreleased', 'blue', attrs=['bold', 'underline']) + '\n')
    for component, entries in _group_entries_by_component(changelog.entries):
      maxw = max(len(', '.join(x.types)) for x in entries)
      fp.write('  ' + colored(component or 'No Component', 'red', attrs=['bold', 'underline']) + '\n')
      for entry in entries:
        lines = textwrap.wrap(entry.description, width - (maxw + 6))
        suffix_fmt = ' '.join(filter(bool, (_fmt_issues(entry), _fmt_components(entry))))
        lines[-1] += ' ' + suffix_fmt
        delta = maxw - len(', '.join(entry.types))
        fp.write('    {} {}\n'.format(colored((_fmt_types(entry) + ':') + ' ' * delta, attrs=['bold']), _md_term_stylize(lines[0])))
        for line in lines[1:]:
          fp.write('    {}{}\n'.format(' ' * (maxw+2), _md_term_stylize(line)))
    fp.write('\n')


def render_changelogs_as_markdown(fp: TextIO, changelogs: List[Changelog]) -> None:

  def _fmt_issue(i):
    if str(i).isdigit():
      return '#' + str(i)
    return i

  def _fmt_issues(entry):
    if not entry.issues:
      return None
    return '(' + ', '.join(_fmt_issue(i) for i in entry.issues) + ')'

  for changelog in changelogs:
    fp.write('## {}\n\n'.format(changelog.version or 'Unreleased'))
    for component, entries in _group_entries_by_component(changelog.entries):
      fp.write('* __{}__\n'.format(component))
      for entry in entries:
        description ='**' + ', '.join(entry.types) + '**: ' + entry.description
        if entry.issues:
          description += ' ' + _fmt_issues(entry)
        lines = textwrap.wrap(description, 80)
        fp.write('    * {}\n'.format(lines[0]))
        for line in lines[1:]:
          fp.write('      {}\n'.format(line))
    fp.write('\n')


def render_changelogs(fp: TextIO, format: str, changelogs: List[Changelog]) -> None:
  changelogs = list(changelogs)
  unreleased = next((x for x in changelogs if not x.version), None)
  if unreleased:
    changelogs.remove(unreleased)
  changelogs.sort(key=lambda x: x.version, reverse=True)
  if unreleased:
    changelogs.insert(0, unreleased)
  Changelog.RENDERERS[format](fp, changelogs)


Changelog.RENDERERS.update({
  'terminal': render_changelogs_for_terminal,
  'markdown': render_changelogs_as_markdown,
})
